/*
** Module   : terminal_win.c
** Abstract : Windows implementation for character mode-based console functions.
**
** Copyright (c) 2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --------------------------------Description-----------------------------------
** 001 EVL 20130729 Created initial version to implement Windows specific code to handle text
**                  console terminal JNI.
*/

#include "terminal.h"
#include <windows.h>

// Defining our own attribite flags, Windows does not have equivalents
#define ATTR_NORMAL     0x00010000
#define ATTR_STANDOUT   0x00020000
#define ATTR_UNDERLINE  0x00040000
#define ATTR_REVERSE    0x00080000
#define ATTR_BLINK      0x00100000
#define ATTR_DIM        0x00200000
#define ATTR_BOLD       0x00400000
#define ATTR_ALTCHARSET 0x00800000
#define ATTR_INVIS      0x01000000

// pointer to the structure of the interactive terminal
static HANDLE hConsoleIn  = NULL;
static HANDLE hConsoleOut = NULL;
static DWORD consoleInModeInitial = 0;
static DWORD consoleOutModeInitial = 0;
static CONSOLE_SCREEN_BUFFER_INFO screenBufferInfoCurrent;
static CONSOLE_CURSOR_INFO cursorInfoCurrent;

static int hascolors = 0;	      // set when we initialize
static int colors_started = 0;	// set when we call start_color().

// This is the normal text attribute may be different depending on console initial color setting
static WORD normalTextAttribute =
   FOREGROUND_BLUE | FOREGROUND_GREEN | FOREGROUND_RED | FOREGROUND_INTENSITY | BACKGROUND_BLUE;
// We need to remember the text attribute to be able to restore it before exiting
#define ATTR_UNKNOWN      0xFFFF
static WORD initialTextAttribute = ATTR_UNKNOWN;

// The array of the character attributes to use in text drawing
// TODO: Check if the constants are correct
jint attribute_result[ATTRIBUTE_MAX] =
   {ATTR_NORMAL,  // A_NORMAL
    ATTR_STANDOUT,  // A_STANDOUT
    ATTR_UNDERLINE,  // A_UNDERLINE
    ATTR_REVERSE,  // A_REVERSE
    ATTR_BLINK,  // A_BLINK ??
    ATTR_DIM,  // A_DIM
    ATTR_BOLD,  // A_BOLD
    ATTR_ALTCHARSET,  // A_ALTCHARSET
    ATTR_INVIS}; // A_INVIS (is black on black invisible?)

// The array of the special symbols to draw the horizontal and vertical lines and other special
// symbols.
// TODO: This is OEM character table entry from 850 CP. And this is seems to be the only way
//       we can work in Windows console, it does not like Unicode.
jint acs_result[ACS_MAX] = {0xDA,  // ASC_ULCORNER
                            0xC0,  // ACS_LLCORNER
                            0xBF,  // ACS_URCORNER
                            0xD9,  // ACS_LRCORNER
                            0xB4,  // ACS_LTEE
                            0xC3,  // ACS_RTEE
                            0xC1,  // ACS_BTEE
                            0xC2,  // ACS_TTEE
                            0xC4,  // ACS_HLINE                                   
                            0xB3,  // ACS_VLINE
                            0xC5,  // ACS_PLUS
                            0xF0,  // ACS_S1 - dash or small minus
                            0x5F,  // ACS_S9 - underscore                           
                            0x04,  // ACS_DIAMOND
                            0xB0,  // ACS_CKBOARD      
                            0xF8,  // ACS_DEGREE
                            0xF1,  // ACS_PLMINUS
                            0x07,  // ACS_BULLET
                            0x18,  // ACS_UARROW
                            0x19,  // ACS_DARROW
                            0x1B,  // ACS_LARROW
                            0x1A,  // ACS_RARROW,
                            0xFB}; // ACS_SELECT
                                                                                    
// Externals definitions                                                            
// terminal size at the moment of initialization of the library 
extern int base_rows;
extern int base_cols;

// This one is implemented in module process_win
void clearScreen(HANDLE);

// ----------------------------------------
// Local functions used in this module only

/**
 * Extracts the attribute value for the given char to draw.
 *
 * @param   charOut     character to draw, upper 16 bits containing the new attribute value.
 *
 * @return   The new attribute to draw the character in Windows compatible format.
 */
WORD decodeAttribute(int charOut)
{
   // Leave only high half of the int
   charOut &= 0xFFFF0000;
   
   // Initially this is the normal attribute for text rendering
   WORD attrResult = normalTextAttribute;

   // Guess the reverse has priority over underline   
   if ((charOut & ATTR_REVERSE) == ATTR_REVERSE)
   {
      attrResult = BACKGROUND_BLUE | BACKGROUND_GREEN | BACKGROUND_RED;
   }
   // The underline attribute is dependent on the current reverse one because in Windows
   // console there is no actual underlined chars support, reverse to current is used instead. 
   if ((charOut & ATTR_UNDERLINE) == ATTR_UNDERLINE)
   {
      if ((charOut & ATTR_REVERSE) == ATTR_REVERSE)
      {
         attrResult = normalTextAttribute;
      }
      else
      {
         attrResult = BACKGROUND_BLUE | BACKGROUND_GREEN | BACKGROUND_RED;
      }
   }
   attrResult |= (charOut & ATTR_BLINK) ? BACKGROUND_INTENSITY : 0;
   attrResult |= (charOut & ATTR_BOLD) ? FOREGROUND_INTENSITY : 0;
   
   return attrResult; 
}

#define WIN_VK_DOWN        0x28
#define WIN_VK_UP          0x26
#define WIN_VK_LEFT        0x25
#define WIN_VK_RIGHT       0x27
#define WIN_VK_HOME        0x24
#define WIN_VK_F1          0x70
#define WIN_VK_F2          0x71
#define WIN_VK_F3          0x72
#define WIN_VK_F4          0x73
#define WIN_VK_F5          0x74
#define WIN_VK_F6          0x75
#define WIN_VK_F7          0x76
#define WIN_VK_F8          0x77
#define WIN_VK_F9          0x78
#define WIN_VK_F10         0x79
#define WIN_VK_F11         0x7a
#define WIN_VK_F12         0x7b
#define WIN_VK_F13         0x7c
#define WIN_VK_F14         0x7d
#define WIN_VK_F15         0x7e
#define WIN_VK_F16         0x7f
#define WIN_VK_F17         0x80
#define WIN_VK_F18         0x81
#define WIN_VK_F19         0x82
#define WIN_VK_F20         0x83       
#define WIN_VK_DELETE      0x2e
#define WIN_VK_INSERT      0x2d
#define WIN_VK_PAGE_DOWN   0x22
#define WIN_VK_PAGE_UP     0x21
#define WIN_VK_END         0x23

#define NC_VK_DOWN       0402
#define NC_VK_UP         0403 
#define NC_VK_LEFT       0404
#define NC_VK_RIGHT      0405
#define NC_VK_HOME       0406
#define NC_VK_F1         0411
#define NC_VK_F2         0412
#define NC_VK_F3         0413
#define NC_VK_F4         0414
#define NC_VK_F5         0415
#define NC_VK_F6         0416
#define NC_VK_F7         0417
#define NC_VK_F8         0420
#define NC_VK_F9         0421
#define NC_VK_F10        0422
#define NC_VK_F11        0423
#define NC_VK_F12        0424
#define NC_VK_F13        0425
#define NC_VK_F14        0426
#define NC_VK_F15        0427
#define NC_VK_F16        0430
#define NC_VK_F17        0431
#define NC_VK_F18        0432
#define NC_VK_F19        0433
#define NC_VK_F20        0434
#define NC_VK_DELETE     0512
#define NC_VK_INSERT     0513
#define NC_VK_PAGE_DOWN  0522
#define NC_VK_PAGE_UP    0523
#define NC_VK_END        0550

/**
 * Converst the key codes gotten from Windows into NCURSES compatible ones. Works for F* and
 * cursor movement keys.
 *
 * @param   key     key code got from Windows console API calls.
 *
 * @return   The code for respecive keys for NCURSES environment.
 */
int transformToNCURSES(int key)
{
   
   switch (key)
   {
      case WIN_VK_DOWN:
         key = NC_VK_DOWN;
         break;
   
      case WIN_VK_UP:
         key = NC_VK_UP;
         break;
   
      case WIN_VK_LEFT:
         key = NC_VK_LEFT;
         break;
   
      case WIN_VK_RIGHT:
         key = NC_VK_RIGHT;
         break;
   
      case WIN_VK_HOME:
         key = NC_VK_HOME;
         break;
   
      case WIN_VK_F1:
         key = NC_VK_F1;
         break;
   
      case WIN_VK_F2:
         key = NC_VK_F2;
         break;
   
      case WIN_VK_F3:
         key = NC_VK_F3;
         break;
   
      case WIN_VK_F4:
         key = NC_VK_F4;
         break;
   
      case WIN_VK_F5:
         key = NC_VK_F5;
         break;
   
      case WIN_VK_F6:
         key = NC_VK_F6;
         break;
   
      case WIN_VK_F7:
         key = NC_VK_F7;
         break;
   
      case WIN_VK_F8:
         key = NC_VK_F8;
         break;
   
      case WIN_VK_F9:
         key = NC_VK_F9;
         break;
   
      case WIN_VK_F10:
         key = NC_VK_F10;
         break;
   
      case WIN_VK_F11:
         key = NC_VK_F11;
         break;
   
      case WIN_VK_F12:
         key = NC_VK_F12;
         break;
   
      case WIN_VK_F13:
         key = NC_VK_F13;
         break;
   
      case WIN_VK_F14:
         key = NC_VK_F14;
         break;
   
      case WIN_VK_F15:
         key = NC_VK_F15;
         break;
   
      case WIN_VK_F16:
         key = NC_VK_F16;
         break;
   
      case WIN_VK_F17:
         key = NC_VK_F17;
         break;
   
      case WIN_VK_F18:
         key = NC_VK_F18;
         break;
   
      case WIN_VK_F19:
         key = NC_VK_F19;
         break;
   
      case WIN_VK_F20:
         key = NC_VK_F20;
         break;
   
      case WIN_VK_DELETE:
         key = NC_VK_DELETE;
         break;
   
      case WIN_VK_INSERT:
         key = NC_VK_INSERT;
         break;
   
      case WIN_VK_PAGE_DOWN:
         key = NC_VK_PAGE_DOWN;
         break;
   
      case WIN_VK_PAGE_UP:
         key = NC_VK_PAGE_UP;
         break;
   
      case WIN_VK_END:
         key = NC_VK_END;
         break;
   }
   return key;
}
/**
 * Clean up function to restore console defaults when usage is finished.
 */
void terminateConsole()
{
   // Restore the text drawing attribute
   if (hConsoleOut != NULL && initialTextAttribute != ATTR_UNKNOWN)
   {
      SetConsoleTextAttribute(hConsoleOut, initialTextAttribute);
   }
   // General system resource deallocation
   FreeConsole();
}
// Local functions used in this module only
// ----------------------------------------

// Public functions start here
/**
 * Windows specific implementation of the moving cursor to the specified position in console.
 * The next character on the console will be placed in this location.
 *
 * @param   yRow     new Y(row) coordinate of the next char to draw.
 *
 * @param   xColumn  new X(column) coordinate of the next char to draw.
 */
void moveNative(int yRow, int xColumn)
{
   // TODO: Implement
   COORD newCoord;
   
   newCoord.X = xColumn;
   newCoord.Y = yRow;
   
   SetConsoleCursorPosition(hConsoleOut, newCoord);
}

/**
 * Windows specific implementation of the character drawing to the current position in console.
 *
 * @param   chrToDraw     character to draw in the current cell.
 */
void addchNative(int chrToDraw)
{
   // TODO: Implement
   static WORD attrCurrent = 0xFFFF;
   // Change the text attribute if required
   WORD attr = decodeAttribute(chrToDraw);

   // 0 attribute is also possible meaning black on black
   if (attrCurrent != attr)
   {
      // Change text attribute when it is really neccessary
      if (!SetConsoleTextAttribute(hConsoleOut, attr))
      {
         // Error, exiting
         return;
      }
      // Save the currently used attribute
      attrCurrent = attr;
   }
   
   // Put the character to console, the char located at low 16 bit of integer
   // (even when we will possibly implent Unicode support)                          
   CHAR chr = chrToDraw & 0x0000FFFF;
   DWORD dwChrWtitten = 0;
   
   WriteConsole(hConsoleOut, &chr, 1, &dwChrWtitten, NULL);
   // This is another way to write the characters to console
/*   CHAR_INFO chri;
   chri.Char.AsciiChar = (unsigned char)(chrToDraw & 0x000000FF);
////   chri.Attributes = decodeAttribute(ATTR_NORMAL);
   chri.Attributes = FOREGROUND_BLUE | FOREGROUND_GREEN | FOREGROUND_RED;
   
   COORD coordBuffSize = {1, 1};
   COORD coordBufCoord = {0, 0};
                                                                                
   CONSOLE_SCREEN_BUFFER_INFO csbInfo;
   GetConsoleScreenBufferInfo(hConsoleOut, &csbInfo);
   
   SMALL_RECT srectWrite;
   srectWrite.Top = csbInfo.dwCursorPosition.Y;
   srectWrite.Bottom = csbInfo.dwCursorPosition.Y;
   srectWrite.Left = csbInfo.dwCursorPosition.X;
   srectWrite.Right = csbInfo.dwCursorPosition.X;
   WriteConsoleOutputA(hConsoleOut, &chri, coordBuffSize, coordBufCoord, &srectWrite);*/
}      

/**
 * Windows specific implementation of the calculating the color according to the given color pair
 * index.  The new attribute returned will be or-ed with the resulting color.
 *
 * @param   attrInitial     initial value of the attribute to work with.
 *
 * @param   colorPair       the color pair index to be mixed with initial attribute value.
 *
 * @return   The mixed attribute value as result of initial attribute and calculated color.
 */
int colorPairNative(int attrInitial, int colorPair)
{
   int attrResult = attrInitial;

   // TODO: Implement, this really can be ignored because we never call init_colorpair() so
   //       there is no valid colorpair known to the system. At this time we do not support
   //       colors
   if (colors_started)
   {
//      attrResult |= COLOR_PAIR(colorPair);
   }
   
   return attrResult;
}

/**
 * Beeps the tone on the terminal.
 */
void beepNative()
{
   // Simple alert beep
   MessageBeep(0xFFFFFFFF);
}

/**
 * Clears the current console screen.
 */
void clearNative()
{
   // Clear terminal screen
   clearScreen(hConsoleOut);
}

/**
 * Returns the safe termname without null string on return.
 *
 * @return   The terminal type or "" on any failure.
 */
char* safe_termname()
{
   // TODO: Implement, what is the default terminal name we will simulate?
   return "vt100";
}

/**
 * Get X coordinate of the cursor.
 *
 * @return   The integer value of the X coordinate.
 */
jint getx()
{
   CONSOLE_SCREEN_BUFFER_INFO csbInfo;
   
   if (GetConsoleScreenBufferInfo(hConsoleOut, &csbInfo))
   {
      return (jint) csbInfo.dwCursorPosition.X;
   }
   else
   {
      return -1;
   }
}

/**
 * Get Y coordinate of the cursor.
 *
 * @param   env     pointer to a structure that provides access to JNI environment.
 *
 * @param   jo      reference to the Java object who called this function.
 *
 * @return   The integer value of the Y coordinate.
 */
jint gety()
{
   CONSOLE_SCREEN_BUFFER_INFO csbInfo;
   
   if (GetConsoleScreenBufferInfo(hConsoleOut, &csbInfo))
   {
      return (jint) csbInfo.dwCursorPosition.Y;
   }
   else
   {
      return -1;
   }
}

/**
 * Check if the terminal is able to display colors or not.
 *
 * @return   <code>true</code> if display is color capable, <code>false</code> otherwise.
 */
jboolean hasColors()
{
// EVL***   return has_colors();
   // TODO: Implement, the Windows console always has colors enabled?
   return hascolors;
}

/**
 * This method performs console initialization on the native side.
 *
 * @param   env     pointer to a structure that provides access to JNI environment.
 *
 */
void initConsole(JNIEnv *env, char * termname)
{
/*
   char *termname;

   termname = getenv("TERM");
    
   if (termname == NULL || *termname == '\0')
   {
      termname = "vt100";
   }
   
   // use this instead of initscr because there is no way to query the
   // current terminal and when you use multiple terms, newterm overwrites
   // the current terminal... so you can never get it if you don't use
   // newterm directly
   terminal = newterm(termname, stdout, stdin);

   if (terminal == NULL)
   {
      char* spec   = "Interactive terminal of type %s cannot be initialized.";
      char* errmsg = malloc(strlen(spec) + strlen(termname) + 1); 
      sprintf(errmsg, spec, termname);
      throwException(env, "java/lang/IllegalStateException", errmsg);
      return;
   }
    
   // ACS_* are not constans and can not be included into static array declarations and should
   // be calculated after new terminal has been created.
   acs_result[0]  = ACS_ULCORNER;
	acs_result[1]  = ACS_LLCORNER;
	acs_result[2]  = ACS_URCORNER;
	acs_result[3]  = ACS_LRCORNER;
	acs_result[4]  = ACS_LTEE;
	acs_result[5]  = ACS_RTEE;
	acs_result[6]  = ACS_BTEE;
	acs_result[7]  = ACS_TTEE;
	acs_result[8]  = ACS_HLINE;
	acs_result[9]  = ACS_VLINE;
	acs_result[10] = ACS_PLUS;
	acs_result[11] = ACS_S1;
	acs_result[12] = ACS_S9;
	acs_result[13] = ACS_DIAMOND;
	acs_result[14] = ACS_CKBOARD;
	acs_result[15] = ACS_DEGREE;
	acs_result[16] = ACS_PLMINUS;
	acs_result[17] = ACS_BULLET;
    
   keypad(stdscr, TRUE);      // enable keyboard mapping
   timeout(500);              // wait up to 500msec for input
   noecho();                  // don't echo input
   raw();
   nonl();
   auto_getch_refresh(FALSE); // disable sync during getch()
   def_prog_mode();           // save our state (after our mode changes)

   hascolors = has_colors();

   getmaxyx(stdscr, base_rows, base_cols);
*/
   // This is failed if we already has the console window
   AllocConsole();
   // Getting the standard handles
   hConsoleIn = GetStdHandle(STD_INPUT_HANDLE);
   hConsoleOut = GetStdHandle(STD_OUTPUT_HANDLE);
   // Set new console modes
   SetConsoleMode(hConsoleIn, 0);
   SetConsoleMode(hConsoleOut, 0);
   // Save modified console modes
   GetConsoleMode(hConsoleIn, &consoleInModeInitial);
   GetConsoleMode(hConsoleOut, &consoleOutModeInitial);
   // Initialize statics
   hascolors = JNI_TRUE;
   CONSOLE_SCREEN_BUFFER_INFO csbInfo;
   if (GetConsoleScreenBufferInfo(hConsoleOut, &csbInfo))
   {
      base_rows = csbInfo.srWindow.Bottom-csbInfo.srWindow.Top+1;
      base_cols = csbInfo.srWindow.Right-csbInfo.srWindow.Left+1;
      initialTextAttribute = csbInfo.wAttributes;
      if (!hascolors)
      {
         normalTextAttribute = initialTextAttribute;
      }
   }
   // Set console title
   SetConsoleTitle("P2J Character Client");
   // And codepage ???
//   SetConsoleOutputCP(850);
//   SetConsoleCP(850);
   // Set up the clean function to safe console termination
   atexit((void (*)(void))terminateConsole);
}

/**
 * This method reads the key from keyboard by the special reading thread from Windows console.
 *
 * @return   The integer code of the key that was read or -1 if there is no character to read.
 */
jint readKey()
{
   INPUT_RECORD inRec;
   DWORD numberOfRecRead = 0;
   
   if (ReadConsoleInput(hConsoleIn, &inRec, 1, &numberOfRecRead))
   {
      if (numberOfRecRead > 0 && inRec.EventType == KEY_EVENT && inRec.Event.KeyEvent.bKeyDown)
      {
         // We need this call to disable event passing to parent console
         if (FlushConsoleInputBuffer(hConsoleIn))
         {
            // If there is a char, get it otherwise get virtual key
            if (inRec.Event.KeyEvent.uChar.AsciiChar != 0)
            {
               return (jint) inRec.Event.KeyEvent.uChar.AsciiChar;
            }
            else
            {
               // Special transformation required here but only for keys without
               // char value defined.
               return (jint) transformToNCURSES(inRec.Event.KeyEvent.wVirtualKeyCode);
            }
         }
      }
   }

   return (jint) -1;
}

/**
 * Reset terminal settings.  May be called only if suspend() was previously called. 
 */
void resetMode()
{
// EVL***   reset_prog_mode();
   // Re allocate console
   AllocConsole();
   // And restore the modes
   SetConsoleMode(hConsoleIn, consoleInModeInitial);
   SetConsoleMode(hConsoleOut, consoleOutModeInitial);
}

/**
 * Re-activates the console mode after ending of the previous suspension state.
 */
void resume()
{
   // Return to the previous tty mode stored locally by suspend() call
   resetMode();
   // Refresh to restore the screen content was before suspension
   SetConsoleScreenBufferSize(hConsoleOut, screenBufferInfoCurrent.dwSize);
   SetConsoleCursorPosition(hConsoleOut, screenBufferInfoCurrent.dwCursorPosition);
   SetConsoleCursorInfo(hConsoleOut, &cursorInfoCurrent);
// EVL***   doupdate();
}

/**
 * Set absolute cursor position
 *
 * @param   x       new X coordinate of the cursor.
 *
 * @param   y       new Y coordinate of the cursor.
 *
 * @return   <code>true</code> if cursor positioning succeeded.
 */
jboolean setCursorNative(jint x, jint y)
{
   // Set current cursor position
   moveNative(y, x);

   // Check if it was OK
   CONSOLE_SCREEN_BUFFER_INFO csbInfo;
   
   if (GetConsoleScreenBufferInfo(hConsoleOut, &csbInfo))
   {
      return csbInfo.dwCursorPosition.Y == y && csbInfo.dwCursorPosition.X == x;
   }
   else
   {
      return JNI_FALSE;
   }
}

/**
 * Set the cursor visibility on or off. This method only affects the visibility of the cursor.
 *
 * @param   env     pointer to a structure that provides access to JNI environment.
 *
 * @param   on      when <code>true</code> the cursor will become visible, when <code>false</code>
 *                  the cursor will become invisible.
 *
 * @return   The previous state of the cursor (<code>true</code> means visible).
 */
jboolean setCursorStatus(JNIEnv *env, jboolean on)
{
/* EVL***   int rc = ERROR;
   jboolean ret;
   
   if( on )
      rc = curs_set( 1 );
   else
      rc = curs_set( 0 );
   
   if ( rc == 1 || rc == 2 )
      ret = JNI_TRUE;
   else if ( rc == 0 )
      ret = JNI_FALSE;
   else
   {
      char* spec;
      char* termname = getenv("TERM");
      if ( rc == ERROR )
         return JNI_TRUE;
      else
         spec = "Unexpected return from curs_set() for terminal of type %s (%d).";
      char* errmsg = malloc(strlen(spec) + strlen(termname) + 32); 
      sprintf(errmsg, spec, termname, rc);
      throwException(env, "java/lang/IllegalStateException", errmsg);
   }
*/   
   jboolean ret = JNI_FALSE;
   
   CONSOLE_CURSOR_INFO ccInfo;
   
   if (GetConsoleCursorInfo(hConsoleOut, &ccInfo))
   {
      ccInfo.bVisible = on;
      if (SetConsoleCursorInfo(hConsoleOut, &ccInfo))
      {
         ret = JNI_TRUE;
      }
   }
   
   return ret;
}

/**
 * Simple reset of terminal type.
 *
 * @param   env       pointer to a structure that provides access to JNI environment.
 *
 * @param   termname  The name of the terminal to be restarted.
 */
void restartTerminal(JNIEnv *env, char* termname)
{
/* EVL***   int   err  = 0;
   int   rc   = restartterm(termname, fileno(stdout), &err); 
   
   if (rc == ERR)
   {
      char* spec   = "Failure resetting terminal of type %s (error %i).";
      char* errmsg = malloc(strlen(spec) + strlen(termname) + 15); 
      sprintf(errmsg, spec, termname, err);
      throwException(env, "java/lang/IllegalStateException", errmsg);
   }*/
   // TODO: Find out what we have to do here if any
}

/**
 * Save the current state of the console mode and switch to TTY mode temporarily.
 */
void suspend()
{
// EVL***   def_prog_mode();            /* Save the tty modes             */
// EVL***   endwin();                   /* End curses mode temporarily    */
   // Save the current console modes
   GetConsoleMode(hConsoleIn, &consoleInModeInitial);
   GetConsoleMode(hConsoleOut, &consoleOutModeInitial);
   // Save screen buffer info
   GetConsoleScreenBufferInfo(hConsoleOut, &screenBufferInfoCurrent);
   // Save cursor info
   GetConsoleCursorInfo(hConsoleOut, &cursorInfoCurrent);
   // Free console
   terminateConsole();
}

/**
 * Synchronizes the state of the console buffers with native terminal window. Meaning the kind of
 * refresh calls depending on the backend used.
 */
void syncNative()
{
// EVL***   refresh();
   fflush(stdout);
}
